@@ -20,7 +20,6 @@
 #include <stdio.h>		/* for printf */
 #include <unistd.h>
 #include <string.h>		/* for memset */
-#include <sys/mman.h>		/* mmap */
 #include "vga.h"
 #include "libvga.h"
 #include "driver.h"
@@ -138,32 +137,34 @@ static int cirrus_chiprev;
 static int cirrus_pci_linear = 0;
 static unsigned char actualMCLK, programmedMCLK;
 static int DRAMbandwidth, DRAMbandwidthLimit;
-static unsigned char *mmio_base;
 
 
 static CardSpecs *cardspecs;
 
-#define NU_FIXED_CLOCKS 19
+#define NU_FIXED_CLOCKS 21
 
 /* 12.588 clock (0x33, 0x3B) replaced with 12.599 (0x2C, 0x33). */
 
 static int cirrus_fixed_clocks[NU_FIXED_CLOCKS] =
 {
-    12599, 25227, 28325, 31500, 36025, 37747, 39992, 41164,
+    12599, 18000, 19600,
+    25227, 28325, 31500, 36025, 37747, 39992, 41164,
     45076, 49867, 64983, 72163, 75000, 80013, 85226, 89998,
     95019, 100226, 108035
 };
 
 static unsigned char fixed_clock_numerator[NU_FIXED_CLOCKS] =
 {
-    0x2C, 0x4A, 0x5B, 0x42, 0x4E, 0x3A, 0x51, 0x45,
+    0x2C, 0x27, 0x28,
+    0x4A, 0x5B, 0x42, 0x4E, 0x3A, 0x51, 0x45,
     0x55, 0x65, 0x76, 0x7E, 0x6E, 0x5F, 0x7D, 0x58,
     0x49, 0x46, 0x53
 };
 
 static unsigned char fixed_clock_denominator[NU_FIXED_CLOCKS] =
 {
-    0x33, 0x2B, 0x2F, 0x1F, 0x3E, 0x17, 0x3A, 0x30,
+    0x33, 0x3e, 0x3a,
+    0x2B, 0x2F, 0x1F, 0x3E, 0x17, 0x3A, 0x30,
     0x36, 0x3A, 0x34, 0x32, 0x2A, 0x22, 0x2A, 0x1C,
     0x16, 0x14, 0x16
 };
@@ -548,8 +549,8 @@ static void cirrus_initializemode(unsigned char *moderegs,
 	/* CRT FIFO threshold setting. */
 	unsigned char threshold;
 	threshold = 8;
-	if (cirrus_chiptype >= CLGD5434)
-	    threshold = 1;
+/*	if (cirrus_chiptype >= CLGD5434)
+	    threshold = 1;*/
 	/* XXX Needs more elaborate setting. */
 	SETBITS(moderegs[CIRRUS_PERFTUNING], 0x0F, threshold);
     }
@@ -1181,12 +1182,6 @@ static int cirrus_init(int force, int par1, int par2)
     __svgalib_cirrus_driverspecs.accelspecs = malloc(sizeof(AccelSpecs));
     __svgalib_clear_accelspecs(__svgalib_cirrus_driverspecs.accelspecs);
     __svgalib_cirrus_driverspecs.accelspecs->flags = ACCELERATE_ANY_LINEWIDTH;
-    if (cirrus_chiptype >= CLGD5429) {
-	/* Map memory-mapped I/O register space. */
-	mmio_base = valloc(32 * 1024);
-	mmap(mmio_base, 32 * 1024, PROT_WRITE, MAP_FIXED | MAP_SHARED,
-	     __svgalib_mem_fd, 0xB8000);
-    }
     /* Set up the correct paging routine */
     if (cirrus_memory >= 2048)
 	__svgalib_cirrus_driverspecs.__svgalib_setpage =
@@ -1194,6 +1189,12 @@ static int cirrus_init(int force, int par1, int par2)
 
     __svgalib_driverspecs = &__svgalib_cirrus_driverspecs;
 
+    __svgalib_banked_mem_base=0xa0000;
+    __svgalib_banked_mem_size=0x10000;
+    __svgalib_linear_mem_base=cirrus_pci_linear;
+    __svgalib_linear_mem_size=cirrus_memory*0x400;
+    __svgalib_mmio_size=32768;
+    __svgalib_mmio_base=0xb8000;
     return 0;
 }
 
@@ -1425,61 +1426,61 @@ static int cirrus_init(int force, int par1, int par2)
 #define MMIOBLTSTATUS		0x40
 
 #define MMIOSETDESTADDR(addr) \
-  *(unsigned int *)(mmio_base + MMIODESTADDR) = addr;
+  *(unsigned int *)(MMIO_POINTER + MMIODESTADDR) = addr;
 
 #define MMIOSETSRCADDR(addr) \
-  *(unsigned int *)(mmio_base + MMIOSRCADDR) = addr;
+  *(unsigned int *)(MMIO_POINTER + MMIOSRCADDR) = addr;
 
 /* Pitch: the 5426 goes up to 4095, the 5434 can do 8191. */
 
 #define MMIOSETDESTPITCH(pitch) \
-  *(unsigned short *)(mmio_base + MMIODESTPITCH) = pitch;
+  *(unsigned short *)(MMIO_POINTER + MMIODESTPITCH) = pitch;
 
 #define MMIOSETSRCPITCH(pitch) \
-  *(unsigned short *)(mmio_base + MMIOSRCPITCH) = pitch;
+  *(unsigned short *)(MMIO_POINTER + MMIOSRCPITCH) = pitch;
 
 /* Width: the 5426 goes up to 2048, the 5434 can do 8192. */
 
 #define MMIOSETWIDTH(width) \
-  *(unsigned short *)(mmio_base + MMIOWIDTH) = (width) - 1;
+  *(unsigned short *)(MMIO_POINTER + MMIOWIDTH) = (width) - 1;
 
 /* Height: the 5426 goes up to 1024, the 5434 can do 2048. */
 
 #define MMIOSETHEIGHT(height) \
-  *(unsigned short *)(mmio_base + MMIOHEIGHT) = (height) - 1;
+  *(unsigned short *)(MMIO_POINTER + MMIOHEIGHT) = (height) - 1;
 
 #define MMIOSETBLTMODE(m) \
-  *(unsigned char *)(mmio_base + MMIOBLTMODE) = m;
+  *(unsigned char *)(MMIO_POINTER + MMIOBLTMODE) = m;
 
 #define MMIOSETBLTWRITEMASK(m) \
-  *(unsigned char *)(mmio_base + MMIOBLTWRITEMASK) = m;
+  *(unsigned char *)(MMIO_POINTER + MMIOBLTWRITEMASK) = m;
 
 #define MMIOSETROP(rop) \
-  *(unsigned char *)(mmio_base + MMIOROP) = rop;
+  *(unsigned char *)(MMIO_POINTER + MMIOROP) = rop;
 
 #define MMIOSTARTBLT() \
-  *(unsigned char *)(mmio_base + MMIOBLTSTATUS) |= 0x02;
+  *(unsigned char *)(MMIO_POINTER + MMIOBLTSTATUS) |= 0x02;
 
 #define MMIOBLTBUSY(s) \
-  s = *(volatile unsigned char *)(mmio_base + MMIOBLTSTATUS) & 1;
+  s = *(volatile unsigned char *)(MMIO_POINTER + MMIOBLTSTATUS) & 1;
 
 #define MMIOSETBACKGROUNDCOLOR(c) \
-  *(unsigned char *)(mmio_base + MMIOBACKGROUNDCOLOR) = c;
+  *(unsigned char *)(MMIO_POINTER + MMIOBACKGROUNDCOLOR) = c;
 
 #define MMIOSETFOREGROUNDCOLOR(c) \
-  *(unsigned char *)(mmio_base + MMIOFOREGROUNDCOLOR) = c;
+  *(unsigned char *)(MMIO_POINTER + MMIOFOREGROUNDCOLOR) = c;
 
 #define MMIOSETBACKGROUNDCOLOR16(c) \
-  *(unsigned short *)(mmio_base + MMIOBACKGROUNDCOLOR) = c;
+  *(unsigned short *)(MMIO_POINTER + MMIOBACKGROUNDCOLOR) = c;
 
 #define MMIOSETFOREGROUNDCOLOR16(c) \
-  *(unsigned short *)(mmio_base + MMIOFOREGROUNDCOLOR) = c;
+  *(unsigned short *)(MMIO_POINTER + MMIOFOREGROUNDCOLOR) = c;
 
 #define MMIOSETBACKGROUNDCOLOR32(c) \
-  *(unsigned int *)(mmio_base + MMIOBACKGROUNDCOLOR) = c;
+  *(unsigned int *)(MMIO_POINTER + MMIOBACKGROUNDCOLOR) = c;
 
 #define MMIOSETFOREGROUNDCOLOR32(c) \
-  *(unsigned int *)(mmio_base + MMIOFOREGROUNDCOLOR) = c;
+  *(unsigned int *)(MMIO_POINTER + MMIOFOREGROUNDCOLOR) = c;
 
 #define MMIOWAITUNTILFINISHED() \
 	for (;;) { \
