@@ -1,6 +1,6 @@
-                       /*  ET6000 driver  Don Secrest  version 0.24 */
-                       /*   mmmapmmmap version  Sept 11 1998        */
-#include <stdlib.h>
+                       /*  ET6000 driver  Don Secrest  version 1.05     */
+                       /*  version using svgalib_pci_find June 10 1999  */
+#include <stdlib.h>    /*  Testing modeline update. */
 #include <stdio.h>
 #include <unistd.h>
 #include <stdarg.h>
@@ -12,12 +12,11 @@
 #include "vga.h"
 #include "libvga.h"
 #include "driver.h"
-/* #include "ramdac/ramdac.h" */
 #include <linux/pci.h>
-
-#ifdef BACKGROUND
-#include "vgabg.h"
-#endif
+#include "timing.h"
+#include "interface.h"
+#include "vgaregs.h"
+#include "vgapci.h"
 
 #ifdef DB6K
 #include <time.h>
@@ -28,6 +27,7 @@ static void prtmodeinfo(vga_modeinfo *m);
 static char *not_written = "ET6000 subroutine %s is not written yet.\n";
 static unsigned char *textstore = 0;
 static unsigned char *vgatextstore = 0;
+void fdumpregs(FILE *fd,unsigned char regs[],int n);
 #define DNOTW(a) fprintf(outf,not_written,a)
 #define DPRT(a) fprintf(outf,a)
 #else
@@ -35,27 +35,32 @@ static unsigned char *vgatextstore = 0;
 #define DPRT(a)
 #endif  /* DB6K */
 
+#ifdef DBG  /* DBG is for debugging the lode to use modelines. */
+FILE *fdm;
+#endif
+
 #define SEG_SELECT 0x3CD
 
-static unsigned long pci_confgspace = 0;
 static unsigned long base1;
 
 /* The following are set by init so that they may be restored to their  */
 /* state in setmode.                                                    */
-static unsigned char pci13std,pci40std,pci42std,pci58std,pci59std=0;
+static unsigned char pci40std,pci42std,pci57std,pci58std,pci59std=0;
+
+#define LMODEMIN 9    /* No mode below this has been tested for linear    */
+#define LMODEMAX 142  /* No mode above this tested.  These should change. */
 
-#define LMODEMIN 9   /* No mode below this has been tested for linear    */
-#define LMODEMAX 22  /* No mode above this tested.  These should change. */
-/* static unsigned char *et6000_frame_ptr =-1; */ /* Mem pointer for linear.*/
-/* This was used by the sane version which mmapped video memory only once */
 static long et6000_mapped_mem = 0;
 
+static CardSpecs *cardspecs;
+static int et6ksav = VGA_MISCOUTPUT + 1;
+static int ET6000_TOTAL_REGS = VGA_MISCOUTPUT + 14;
 static char mem_type;
 static int et6000_memory;
-static int et6000_chiptype;
+static int et6000_linear_base;
 static int et6000_interlaced(int mode);
-static int et6000_setlinear(int addr);
-static et6000_chiptype = 0;
+static unsigned comp_lmn(unsigned clock,float fac);
+static int et6000_chiptype = 0;
 static int x3_4,x3_5,x3_8 = 0,x3_a; /* On init these get set according */
 				    /*  to color */
 union longword {
@@ -85,45 +90,10 @@ static void et6000_getmodeinfo(int mode,vga_modeinfo *modeinfo);
 static int  et6000_ext_set(unsigned what, va_list params);
 static int  et6000_accel(unsigned operation, va_list params);
 static int  et6000_linear(int op, int param);
+
 /*
 ** End of the DriverSpecs function declarations         
 */
-/* Mode table */
-#if defined(DYNAMIC)
-
-static ModeTable *et6000_modes = NULL;
-static ModeTable No_Modes = END_OF_MODE_TABLE;
-
-#else				/* !defined(DYNAMIC) */
-
-#include "et6000.regs"
-
-static ModeTable et6000_modes[] =
-{
-/* *INDENT-OFF* */
-    OneModeEntry(320x200x32K),
-    OneModeEntry(320x200x64K),
-    OneModeEntry(320x200x16M),
-    OneModeEntry(640x480x256),
-    OneModeEntry(640x480x32K),
-    OneModeEntry(640x480x64K),
-    OneModeEntry(640x480x16M),
-    OneModeEntry(800x600x16),
-    OneModeEntry(800x600x256),
-    OneModeEntry(800x600x32K),
-    OneModeEntry(800x600x64K),
-    /*    OneModeEntry(800x600x16M),   */
-    /*    OneModeEntry(1024x786x32K),  */
-    /*    OneModeEntry(1024x786x64K),  */
-    /*    OneModeEntry(1280x1024x256), */
-    OneModeEntry(1024x768x16),
-    OneModeEntry(1024x768x256),
-    OneModeEntry(1280x1024x16),
-    END_OF_MODE_TABLE
-/* *INDENT-ON* */
-};
-
-#endif				/* !defined(DYNAMIC) */
 
 /* Function table (exported) */
 
@@ -154,6 +124,133 @@ DriverSpecs __svgalib_et6000_driverspecs =
     NULL,			/* accelspecs */
     NULL,			/* emulation */
 };
+#ifdef DBG
+/* programs to debug the modeline stuff. */
+static void pmodeinfo(ModeInfo *In)
+{
+  fprintf(fdm,"Modeinfo\nWidth %d  height %d\n",In->width,In->height);
+  fprintf(fdm,"Bpp %d,  bpp %d,  cB %d\n",In->bytesPerPixel,In->bitsPerPixel,
+	 In->colorBits);
+  fprintf(fdm,"RGB %d,  %d,  %d\n",In->redWeight,In->greenWeight,
+	  In->blueWeight);
+  fprintf(fdm,"Offset RGB %d,  %d,  %d\n",In->redOffset,In->greenOffset,
+	  In->blueOffset);
+  fprintf(fdm,"Line width %d,  real width %d,  real height %d\n",In->lineWidth,
+	  In->realWidth,In->realHeight);
+}
+static void pmodetiming(ModeTiming *Tg)
+{
+  int i,*mt;
+  mt = (int *) Tg;
+  fprintf(fdm,"Mode timing\n");
+  for(i = 0;i < 20;i++)
+    {
+      fprintf(fdm,"0x%04x  ",*(mt + i));
+	if((i+1)%5 == 0) fprintf(fdm,"\n");
+    }
+}
+#endif  /* Modeline routine debug. */
+
+static void et6000_initializemode(unsigned char *moderegs, 
+				  ModeTiming *modetiming,ModeInfo *modeinfo,
+				  int mode)
+{
+  int tmp,tmptot,tmpss,tmpse,tmpbs,tmpbe,offset,i;
+  float fac;
+
+  fac = 1.0;
+  __svgalib_setup_VGA_registers(moderegs,modetiming,modeinfo);
+  offset = modeinfo->lineWidth >>3;
+  if(modetiming->flags & DOUBLESCAN)
+    moderegs[VGA_CR17] = 0xa3;
+  else
+    moderegs[VGA_CR17] = 0xab;
+  if((modeinfo->bytesPerPixel == 2) || (modetiming->flags & DOUBLESCAN))
+      moderegs[VGA_CR14] = 0x40;
+  else
+      moderegs[VGA_CR14] = 0x60;
+  moderegs[VGA_CR13] = offset & 0xff;
+  moderegs[VGA_AR10] = 0x01;     /* This is correct. May 20 1999 */
+  moderegs[VGA_AR11] = 0x00;
+  for(i = 0;i < 13;i++)
+    moderegs[et6ksav + i] = 0x00;
+
+  tmp    = (modetiming->CrtcHDisplay >> 3) - 1;
+  tmptot = (modetiming->CrtcHTotal >> 3) -5;
+  tmpss  = (modetiming->CrtcHSyncStart >> 3);
+  tmpbs  = (modetiming->CrtcHSyncStart >> 3) - 1;
+  tmpse  = offset;                                 /*    3F      */
+  moderegs[et6ksav+8] = ((tmptot & 0x100) >> 8) |  /* 0 Htot (8) */
+                        ((tmp  &  0x100)  >> 7) |  /* 1 HDsp (8) */
+                        ((tmpbs &  0x100) >> 6) |  /* 2 HBs  (8) */
+                        ((tmpss &  0x100) >> 4) |  /* 4 HSs  (8) */
+                        ((tmpse &  0x200) >> 3) |  /* 6 offs (9) */
+                        ((tmpse &  0x100) >> 1);   /* 7 ofss (8) */
+  tmp    = modetiming->CrtcVDisplay -1;
+  tmptot = modetiming->CrtcVTotal - 2;
+  tmpbs  = modetiming->CrtcVSyncStart -1;
+  tmpss  = modetiming->CrtcVSyncStart;
+  tmpbe  = 0x00;  /* (10) */                        /*     35      */
+  moderegs[et6ksav+5] = ((tmpbs & 0x400)  >> 10) |   /* 0 VBs  (10) */
+                        ((tmptot & 0x400) >> 9 ) |   /* 1 Vtot (10) */
+                        ((tmp   &  0x400) >> 8 ) |   /* 2 VDsp (10) */
+                        ((tmpss &  0x400) >> 7 ) |   /* 3 Vsst (10) */
+                        ( tmpbe >> 6);               /* 4 splt (10) */
+  if(modetiming->flags & INTERLACED)
+    moderegs[et6ksav + 5] |= 0xa0;  /* bit 7 of cr35 and bit 1 of 57 */
+
+  switch(modeinfo->bitsPerPixel)
+  {
+  case 8:
+    moderegs[et6ksav + 12] = 0x00;        /* et6ksav   Reg   */
+    moderegs[et6ksav +  6] = 0x04;        /*    0     PCI 58 */
+    moderegs[et6ksav] &= 0xfe;            /*    1     PCI 42 */
+    break;                                /*    2     CR18   */
+  case 15:                                /*    3     CR33   */
+  case 16:                                /*    4     CR34   */
+    moderegs[et6ksav + 12] = 0x90;        /*    5     CR35   */
+    moderegs[et6ksav +  6] = 0x08;        /*    6     PCI 59 */
+    if(modeinfo->greenWeight == 5)        /*    7     AR17   */
+      moderegs[et6ksav] = 0x00;           /*    8     CR3F   */
+    else                                  /*    9     PCI 40 */
+      moderegs[et6ksav] = 0x02;           /*   10     PCI 13 */
+    break;                                /*   11     3CD    */
+  case 24:                                /*   12     AR16   */
+    moderegs[et6ksav + 12] = 0xa0;
+    moderegs[et6ksav  + 9] = 0x06;
+    moderegs[et6ksav  + 6] = 0x02;
+    moderegs[et6ksav  + 2] = 0xff;
+    moderegs[et6ksav  + 1] = 0x02;
+    fac = 3.0;
+    break;
+  case 32:
+    moderegs[et6ksav + 12] = 0xb0;
+    moderegs[et6ksav  + 9] = 0x06;
+    moderegs[et6ksav +  6] = 0x02;
+    moderegs[et6ksav  + 2] = 0xff;
+    moderegs[et6ksav +  1] = 0x03;
+    fac = 4.0;
+    break;
+  default:
+    moderegs[et6ksav + 12] = 0x00;
+    moderegs[et6ksav +  6] = 0x04;
+    moderegs[et6ksav] &= 0xfe;
+    break;
+  }
+
+  modetiming->selectedClockNo=6; /* use clock 6 */
+
+  if(modetiming->selectedClockNo > 3)
+    moderegs[et6ksav + 4] |= 2;
+  moderegs[VGA_MISCOUTPUT] = (moderegs[VGA_MISCOUTPUT] & 0xf3) |
+                             ((modetiming->selectedClockNo & 0x03) << 2);
+   tmp=comp_lmn(modetiming->pixelClock,fac);
+   outb(base1 | 0x68,6);    
+   outb(base1 | 0x69,tmp&0xff);
+   outb(base1 | 0x69,tmp>>8);
+  /* because timing is giving the wrong */
+  /* clock. FIX THIS WHEN CLOCK IS OK.  */
+}
 
 static int  et6000_saveregs(unsigned char regs[]) /* Called frm vga_saveregs */
 {
@@ -165,7 +262,6 @@ static int  et6000_saveregs(unsigned char regs[]) /* Called frm vga_saveregs */
     fprintf(outf,"In saveregs, regs at %x\n",(int) regs);
     fflush(outf);
   }
-  /* return(0);       Testing */
 #endif
   /* Save extended CRT registers. */
   for(i = 0; i < 3;i++) {
@@ -178,7 +274,7 @@ static int  et6000_saveregs(unsigned char regs[]) /* Called frm vga_saveregs */
   regs[EXT + 8] = port_in(__svgalib_CRT_D);
   port_out(0x18,__svgalib_CRT_I);
   regs[EXT + 2] = port_in(__svgalib_CRT_D);
-                                        /* ET6000 only.                    */
+                                        /* ET6000 only.               */
   /* Extended sequencer register 7 in EXT+9 doesen't exist on et6000. */
   /* Also EXT+10 doesn't exist. There is no 0x3c3 on et6000.          */
   regs[EXT + 11] = port_in(SEG_SELECT);
@@ -209,6 +305,7 @@ static int  et6000_saveregs(unsigned char regs[]) /* Called frm vga_saveregs */
   __svgalib_delay();
   regs[EXT + 7] = port_in(ATT_R);
   port_out(0x06,SEQ_I);
+  regs[EXT + 5] = ((inb(base1 | 0x57) & 2) << 4); /* bit 1 of 57 to bit5  */
   regs[EXT + 9] = inb(base1 | 0x40);   /* Sets memory mapping.            */
   regs[EXT + 1] = inb(base1 | 0x42);   /* Saved for future use.           */
   return(13);                          /* ET6000 requires 13 EXT regs.    */
@@ -252,6 +349,7 @@ static void et6000_setregs(const unsigned char regs[], int mode)
   outb(base1 | 0x58,(vga_getcolors() == (1 << 16))? (regs[EXT] | 0x02):
                                                         regs[EXT]);
   outb(base1 | 0x59,regs[EXT + 6]);
+  outb(base1 | 0x57,((inb(base1 | 0x57) | ((regs[EXT + 5] & 0x20) >> 4))));
 
     /* For True Color modes we must devide the MCLK by 3. */
     outb(base1 | 0x42,regs[EXT + 1]);
@@ -308,19 +406,33 @@ static int  et6000_test(void)
     if(et6000_chiptype)
       return(1);
     return(0);
-
     }
 }
+/* Routines needed for modline calculation. */
+static int et6000_map_clock(int bpp, int clock)
+{
+  return(clock);
+}
+
+static int et6000_match_programable_clock(int clock)
+{
+  return(clock);
+}
+
+static int et6000_map_horizontal_crtc(int bpp, int pixelclock, int htiming)
+{
+  return(htiming);
+}
+/* End of modeline subroutines. */
 
 static int  et6000_init(int force, int par1, int par2)
 {
-  int typ,cn;
-  unsigned char pci1,mem_cfg_reg,pci2;
-  unsigned long pcil1,pcil2,prefix;
-  /* unsigned long pm; */ /* For mmaping linear once and for all. */
-  union longword value;
+  unsigned char mem_cfg_reg;
+  unsigned long buf[64];  /* 256 bytes of the pci bus registers. */
+
 #ifdef DB6K
   char dboutfile[30];
+  if(et6000_chiptype) return(1);
   sprintf(dboutfile,"D%ld.txt",time(NULL));
   outf = fopen(dboutfile,"w");
   if(!outf){
@@ -328,122 +440,41 @@ static int  et6000_init(int force, int par1, int par2)
     exit(1);
   }
 #endif
+
   if(et6000_chiptype)
     return(1);             /* An attempt to reenter init. */
-#ifdef DYNAMIC
-    if (et6000_modes == NULL) {
-	FILE *regs;
-
-	regs = fopen(ET6000_REGS, "r");
-	if (regs != 0) {                /* We have a builtin dac. Clocks    */
-	    et6000_modes = NULL;        /* are not used at present.         */
-	    __svgalib_readmodes(regs, &et6000_modes, 0,0);
-	    fclose(regs);
-	} else
-	    et6000_modes = &No_Modes;
-    }
-#endif
   /* Check for ET6000 chiptype.                                             */
   /* Get PCI and VL configuration space.  The VL bus has not been checked. */
   /* I have a PCI bus.                                                     */
-  if(!x3_8) {
+
     /* Test color mode or mono mode */
-    if(inb(0x3cc) & 1) {
-      x3_4 = 0x3d4;
-      x3_5 = 0x3d5;
+    x3_4 = __svgalib_CRT_I;       /* set in vga.c */
+    x3_5 = __svgalib_CRT_D;
+    x3_a = __svgalib_IS1_R;
+    if(__svgalib_CRT_I == 0x3d4)
       x3_8 = 0x3d8;
-      x3_a = 0x3da;
-    }
-    else {
-      x3_4 = 0x3b4;
-      x3_5 = 0x3b5;
+    else
       x3_8 = 0x3b8;
-      x3_a = 0x3ba;
-    }
-  }
+
     if(iopl(3) == -1)
       {
           printf("No permission to use PCI ports.\n");
           perror("PCI ports");
       }
-                              /* We need permission to use ports above 0x3ff */
-                              /* so iopl must be executed while we are       */
-                              /* running as root.  We are first called by    */
-                              /* vga_init and after returning our uid is     */
-                              /* reset to user, so we must leave this        */
-                              /* permission in effect.                       */
-    outb(0xCF8,0);
-    outb(0xCFA,0);
-    pci1 = inb(0xCF8);
-    pci2 = inb(0xCFA);
-                             /* It appears that the byte activation of 0xCF8
-			     ** is for a typ 2 PCI bus or perhaps a VESA bus,
-                             ** I dont know.  My bus is type 1 so I didn't do
-                             ** anything but test it.  I get back 0xFF, so then
-                             ** I try the long iout which works for my bus.
-                             */
-#ifdef DB6K
-fprintf(outf,"0xCF8 gives %d and 0xCFA gives %d\n",pci1,pci2);
-#endif
-    if((pci1 == 0x00) && (pci2 == 0x00))
-      typ = 2;
-    else
-      {
-	pcil1 = inl(0xCF8);
-	outl(0xCF8,0x80000000);
-	pcil2 = inl(0xCF8);
-	outl(0xCF8,pcil2);                  /* Restore CF8 */
-#ifdef DB6K
-fprintf(outf,"0xCF8 long gives %x, the old value was %x\n",(int) pcil2,
-	(int) pcil1);
-#endif
-        if(pcil2 == 0x80000000)
-	  typ = 1;
-	else
-	  typ = 0;
-      }
-    if(typ != 1)
-      {
-	printf("PCI bus error.  We only handle type 1 PCI bus "
-	       "at present.\nThis is type %d\n",typ);
-	return(1);
-      }
-                         /* Here I am assuming type one and searching for the
-			 ** video card.  The search command is a 16 bit number
-                         ** starting with a 1 in the highorder position.  The
-                         ** form is
-			 **        10000000000000nncccccfffoooooooo
-                         ** where
-                         ** nn is the PCI bus number. I assumed only 1 nn = 0
-			 ** ccccc is the card number.  Mine is 0x14.
-			 ** fff is the function.  Mine is 0 and I assume it is
-			 ** true for video cards. It may not be, I don't know.
-			 ** oooooooo  is the configuration space off set.
-			 */
-    cn = 0;
-    while(cn < 0x20)
-      {
-	prefix = 0x80000000 + (cn << 11);
-	    outl(0xCF8,prefix);
-	    value.l = inl(0xCFC);
-	    if(value.w[0] == 0x100c && value.w[1] == 0x3208)
-	      {
-		          /* Found et6000 
-		          ** The first long word is ddddvvvv where vvvv is the
-			  ** vendor and dddd is the device.  Since the 80x86 is
-			  ** a little endian machine the low order bit is 
-			  ** stored lowest in memory, so word 0 is the vendor 
-			  ** 0x100c Tseng Labs and word 1 is the device 0x3208
-			  **  the et6000 video card.
-			  */
+    et6000_linear_base = 0;
+		if(__svgalib_pci_find_vendor_vga(0x100c,buf) == 0 &&
+		   (buf[0] == 0x3208100c))
+		  {
 DPRT("Found et6000\n");
 		et6000_chiptype = ET6000;   /* Defined in vga.h */
-		pci_confgspace = prefix;
+		  }
+		else return(0);
+
 		et6000_unlock();            /* Set key. */
 		mem_type = inb(0x3c2) & 1;  /* 0 = DRAM, 1 = MDRAM */
 
-                outl(0xCF8,prefix | 0x14);  /* The value base1 is at offset */
-                base1 = inl(0xCFC) & ~0xFF; /* 14 hex. The start of confg.  */
+                base1 = buf[0x5] & ~0xFF; 
+		              /* 14 hex. The start of confg.  */
 		              /* 3C2 is the input status register.  It has
 			      ** things like the monitor ID and the video
 			      ** memory type.  The two low order bits are
@@ -455,13 +486,24 @@ DPRT("Found et6000\n");
 			      */
 		if(!pci59std) {
 		  /* Saving pcibus registers so they may be restored. */
-		  pci13std = inb(base1 | 0x13);
-		  pci40std = inb(base1 | 0x40);
-		  pci42std = inb(base1 | 0x42);
-		  pci58std = inb(base1 | 0x58);
-		  pci59std = inb(base1 | 0x59);
+		  pci40std = (unsigned char) buf[16] & 0xFF;
+		  pci42std = (unsigned char) ((buf[16] & 0xFF0000) >> 16);
+		  pci57std = (unsigned char) ((buf[21] & 0xFF000000) >> 24);
+		  pci58std = (unsigned char) buf[22] & 0xFF;
+		  pci59std = (unsigned char) ((buf[22] & 0xFF00) >> 8);
 		}
-		mem_cfg_reg = inb(base1 | 0x45) & 3;
+		mem_cfg_reg = (unsigned char) ((buf[17] & 0xFF00) >> 8) & 3;
+#ifdef DB6K
+		{
+		  int i;
+		  for(i = 0;i < 64;i+=4)
+			fprintf(outf,"%08lx     %08lx  %08lx  %08lx  %08lx\n",
+				(i/4) << 4,buf[i+3],buf[i+2],buf[i+1],buf[i]);
+		  fprintf(outf,"\npci40std 0x%02x  pci42std 0x%02x  pci58std "
+                  "0x%02x  pci59std 0x%02x\n",pci40std,
+			  pci42std,pci58std,pci59std);
+		}
+#endif
 		if(mem_type)
 		  {   /* MDRAM type */
 		    if(mem_cfg_reg == 1)
@@ -478,31 +520,52 @@ DPRT("Found et6000\n");
 		    else
 		      et6000_memory = 1;      /* 1 MByte or DRAM  */
 		  }
-#if 0	/* Mmap for the frame pointer for linear. */
-		et6000_frame_ptr = (unsigned char *) mmap(
-							  (caddr_t) 0,
-						    et6000_memory*1024*1024,
-						    PROT_READ | PROT_WRITE,
-							  MAP_SHARED,
-							  __svgalib_mem_fd,
-							  0x10000000
-							  );
-#ifdef DB6K
-		fprintf(outf,"Et6k_frame_ptr = %x to %x\n",et6000_frame_ptr,
-			et6000_frame_ptr + et6000_memory*1024*1024);
-fprintf(outf,"pci_confgspace %08x, memtype %d and memsize %d MBytes\n",
-       (int) pci_confgspace,mem_type,et6000_memory);
-#endif
-#endif /* If 0 We need to include some more stuff here to map once. */
-       /* see listing 7/10/98 pencil corrections. */
-                printf("Using Tseng ET6000 driver (%d MBytes %sDRAM)\n",
+    et6000_linear_base = buf[4] & 0xff000000;
+    __svgalib_linear_mem_size = et6000_memory*0x100000;
+    __svgalib_linear_mem_base = et6000_linear_base;
+    /* Banked memory is set in vga.c */
+    outb(base1 | 0x68,0xa);   /* Set clocka */
+    outb(base1 | 0x69,0x24);
+    outb(base1 | 0x69,0x21);
+
+    if(__svgalib_driver_report)
+      printf("Using Tseng ET6000 driver (%d MBytes %sDRAM)\n",
 		       et6000_memory,mem_type ? "M" : "");
-		__svgalib_driverspecs = &__svgalib_et6000_driverspecs;
-                break;  
-	      }
-	    else
-	cn++;
-      }
+    __svgalib_driverspecs = &__svgalib_et6000_driverspecs;
+    cardspecs = malloc(sizeof(CardSpecs));
+    cardspecs->videoMemory = et6000_memory << 10;
+    cardspecs->maxPixelClock4bpp = 75000;
+    cardspecs->maxPixelClock8bpp = 160000;
+    cardspecs->maxPixelClock16bpp = 160000;
+    cardspecs->maxPixelClock24bpp = 160000;
+    cardspecs->maxPixelClock32bpp = 160000;
+    cardspecs->flags = INTERLACE_DIVIDE_VERT | CLOCK_PROGRAMMABLE;
+    cardspecs->maxHorizontalCrtc = 2040;
+    cardspecs->nClocks = 0;
+    cardspecs->clocks = NULL;
+    cardspecs->mapClock = et6000_map_clock;
+    cardspecs->matchProgrammableClock = et6000_match_programable_clock;
+    cardspecs->mapHorizontalCrtc = et6000_map_horizontal_crtc;
+    et6000_lock();
+#ifdef DBG
+    {
+      int i,*csp;
+      fdm = fopen("regset.txt","w");
+      if(!fdm)
+	{
+	  printf("Can't open register setting file.\n");
+	  exit(1);
+	}
+      fprintf(fdm,"CardSpecs\n");
+      csp = (int* ) cardspecs;
+      for(i = 1;i < 13; i++)
+	{
+	  fprintf(fdm,"%d,  ",*(csp+i));
+	  if(i % 3 == 0) fprintf(fdm,"\n");
+	}
+      fprintf(fdm,"videoMemory %d\n",cardspecs->videoMemory);
+    }
+#endif  /* DBG */
     return(0);
 }
 
@@ -540,8 +603,9 @@ static void et6000_setwrpage(int page)
 static int  et6000_setmode(int mode, int prv_mode)
 {
 
-    const unsigned char *regs;
-    unsigned char i;
+    unsigned char i,*moderegs;
+    ModeTiming *modetiming;
+    ModeInfo *modeinfo;
 
 #ifdef DB6K
     fprintf(outf,"In setmode, mode = %d, prv_mode = %d\n",mode,prv_mode);
@@ -561,11 +625,12 @@ DPRT("STDVGADRV\n");
 	    outb(__svgalib_CRT_D, (i & 0xFD));  /* Turn it off. */
 	                              /* Make sure pci bus is set right. */
 #ifdef DB6K
-	fprintf(outf,"reg 42 = %2x before.\n",inb(base1 | 0x42));
+	fprintf(outf,"reg 42 = %2x  58 = %2x before.\n",inb(base1 | 0x42)
+		,inb(base1 | 0x58));
 #endif
-	outb(base1 | 0x13,pci13std);
 	outb(base1 | 0x40,pci40std);
 	outb(base1 | 0x42,pci42std);
+	outb(base1 | 0x57,pci57std);
 	outb(base1 | 0x58,pci58std);
 	outb(base1 | 0x59,pci59std);
 #ifdef DB6K
@@ -576,29 +641,42 @@ DPRT("STDVGADRV\n");
 	return __svgalib_vga_driverspecs.setmode(mode, prv_mode);
     case SVGADRV:
 DPRT("SVGADRV\n");
-	regs = LOOKUPMODE(et6000_modes, mode);
-	if (regs != NULL)
 	    break;
     default:
 	return 1;		/* mode not available */
     }
-    __svgalib_setregs(regs);
-    et6000_setregs(regs, mode);
-    return 0;
+    modeinfo = __svgalib_createModeInfoStructureForSvgalibMode(mode);
+    modetiming = malloc(sizeof(ModeTiming));
+    if(__svgalib_getmodetiming(modetiming,modeinfo,cardspecs))
+      {
+	free(modetiming);
+	free(modeinfo);
+	return(1);
+      }
+    moderegs = malloc(ET6000_TOTAL_REGS);
+    et6000_initializemode(moderegs,modetiming,modeinfo,mode);
+#ifdef DBG
+    fprintf(fdm,"In setmode.\n");
+    pmodetiming(modetiming);
+    fdumpregs(fdm,moderegs,ET6000_TOTAL_REGS);
+#endif
+	__svgalib_setregs(moderegs);
+	et6000_setregs(moderegs,mode);
+	free(moderegs);
+	free(modeinfo);
+	return(0);
 }
 
 static int  et6000_modeavailable(int mode)
 {
-  const unsigned char *regs;
+  int rtn;
   struct info *info;
-  regs = LOOKUPMODE(et6000_modes,mode);
-#ifdef DB6K
-  if(regs)
-    fprintf(outf,"In modeavailable, regs = %x, mode = %d\n",(int) regs,mode);
-#endif
-  if(regs == NULL || mode == GPLANE16)
+  ModeTiming *modetiming;
+  ModeInfo *modeinfo;
+
+  if(mode == G720x348x2 || mode == GPLANE16 || mode < G640x480x256 )
     return(__svgalib_vga_driverspecs.modeavailable(mode));
-  if(regs == DISABLE_MODE || mode <= TEXT || mode > GLASTMODE)
+  if(mode <= TEXT || mode > GLASTMODE)
     return(0);
   info = &__svgalib_infotable[mode];
   if(et6000_memory*1024*1024 < info->ydim*info->xbytes)
@@ -606,7 +684,30 @@ static int  et6000_modeavailable(int mode)
                      ** The DAC is built in and I should be able to create
 		     **	anything I have memory for.  I will DISABLE_MODE what
 		     ** I can't handle. */
-  return(SVGADRV);
+  modeinfo = __svgalib_createModeInfoStructureForSvgalibMode(mode);
+  modetiming = malloc(sizeof(ModeTiming));
+#ifdef DBG
+  fprintf(fdm,"Calling getmodetiming\n");
+#endif
+  if(__svgalib_getmodetiming(modetiming,modeinfo,cardspecs))
+    rtn = 0;
+  else if(modetiming->flags & INTERLACED)
+      rtn = 0;
+  else if(modetiming->flags & DOUBLESCAN && (modeinfo->bytesPerPixel == 1))
+    rtn = 0;
+  else if(modeinfo->bitsPerPixel != 4 
+	  && modeinfo->width != 1152)
+    rtn = SVGADRV;
+  else
+    rtn = 0;
+#ifdef DBG
+  pmodeinfo(modeinfo);
+  pmodetiming(modetiming);
+  fprintf(fdm,"Mode %d %s.\n",mode,rtn?"OK":"failure");
+#endif
+  free(modetiming);
+  free(modeinfo);
+  return(rtn);
 }
 
 static void et6000_setdisplaystart(int address)
@@ -663,18 +764,7 @@ static void et6000_getmodeinfo(int mode,vga_modeinfo *modeinfo)
 #ifdef DB6K
       fprintf(outf,"flag = 0x%x\n",modeinfo->flags);
 #endif
-      if(modeinfo->flags & IS_LINEAR)
-#ifdef BACKGROUND
-      modeinfo->linear_aperture = __svgalib_linearframebuffer;
-#else
-      modeinfo->linear_aperture = (char *) et6000_mapped_mem;
-      /* Change to et6000_frame_ptr for map once. */
-#endif
-      else
-	modeinfo->linear_aperture = 0;
-      /* et6000_frame_ptr; */
     }
-    else modeinfo->linear_aperture = 0;
   modeinfo->chiptype = et6000_chiptype;            /* is set in vgamisc.c  */
   modeinfo->linewidth_unit = 8;
   modeinfo->aperture_size = et6000_memory*1024;
@@ -751,21 +841,14 @@ static int  et6000_accel(unsigned operation, va_list params)
 static int  et6000_linear(int op, int param)
 {
 #ifdef DB6K
-  fprintf(outf,"linear op = %d, param = ox%x ",op,param);
-#endif
-#if 0
-  if(op == LINEAR_QUERY_BASE)
-    if((vga_getmodeinfo(vga_getcurrentmode())->flags | CAPABLE_LINEAR) &&
-       (et6000_frame_ptr != (char *) -1))
-      return((int) et6000_frame_ptr);
+  fprintf(outf,"linear op = %d, param = 0x%0x ",op,param);
 #endif
   if(op == LINEAR_QUERY_BASE) {
 #ifdef DB6K
     fprintf(outf, "QUERY\n");
 #endif
-    if((vga_getmodeinfo(vga_getcurrentmode())->flags | CAPABLE_LINEAR) &&
-       param < 0x70)
-      return((0x7f - param) << 24);
+    if((vga_getmodeinfo(vga_getcurrentmode())->flags | CAPABLE_LINEAR))
+      return(et6000_linear_base);
     else                    /* frame_ptr was determined at init time */
       return(-1);
   }
@@ -775,7 +858,6 @@ static int  et6000_linear(int op, int param)
       fprintf(outf,"ENABLE, param = 0x%x\n",param);
 #endif
       et6000_mapped_mem = param;
-      et6000_setlinear(param >> 24);
       port_out(0,SEG_SELECT);
       port_out(0,0x3cb);
       port_out(5,0x3ce);
@@ -792,10 +874,9 @@ static int  et6000_linear(int op, int param)
 #endif
       outb(base1 | 0x40, 0);
       /* Turn off memory relocation and linear. */
-      et6000_setlinear(0);
       __svgalib_modeinfo_linearset = 0;  /* In vga.c */
       /*      port_out(5,0x3ce);
-      port_out(0x00,0x3cf); */
+      port_out(0x00,0x3cf);   See comment on LINEAR_ENABLE */
       return(0);
     }
     else if(op == LINEAR_QUERY_RANGE) {
@@ -821,7 +902,9 @@ static int  et6000_linear(int op, int param)
 
 static int et6000_interlaced(int mode)
 {
-  const unsigned char *regs;
+  ModeTiming *modetiming;
+  ModeInfo   *modeinfo;
+  int        rtn;
 
   DPRT("In interlaced\n");
 #ifdef DB6K
@@ -829,20 +912,57 @@ static int et6000_interlaced(int mode)
 #endif
     if (et6000_modeavailable(mode) != SVGADRV)
 	return 0;
-    regs = LOOKUPMODE(et6000_modes, mode);
-    if (regs == NULL || regs == DISABLE_MODE)
-	return 0;
-    return (regs[EXT + 5] & 0x80) != 0;		/* CRTC 35H */
-
-                      /* This is a copy from et4000. */
+    modeinfo = __svgalib_createModeInfoStructureForSvgalibMode(mode);
+    modetiming = malloc(sizeof(ModeTiming));
+    if(__svgalib_getmodetiming(modetiming,modeinfo,cardspecs))
+       rtn = 0;
+    else
+      rtn = modeinfo->flags & INTERLACED;
+#ifdef DB6K
+    fprintf(outf,"Interlaced %s\n",rtn?"Yes":"No");
+#endif
+    free(modetiming);
+    free(modeinfo);
+    return(rtn);
 }
 
-static int et6000_setlinear(int addr)
-{
-  et6000_unlock();
-  if(addr)
-    outb(base1 | 0x13,addr);
-  else
-    outb(base1 | 0x13,pci13std);
-  return(0);
+#define WITHIN(v,c1,c2) (((v) >= (c1)) && ((v) <= (c2)))
+
+static unsigned comp_lmn(unsigned clock,float fac){
+  int     m, n1, n2;
+  double  fvco;
+  double  fout;
+  double  fmax, fmin;
+  double  fref;
+  double  fvco_goal;
+
+  fmax = 230000.0;
+  fmin = 48000.0;
+
+/* fmin, fmax and limits for n and m are guesses */
+
+  fref = 14318; 
+
+  for (n1 = 4; n1 <= 33; n1++)
+  {
+    for (n2 = 0; n2 <= 3; n2++)
+    {
+      for (m = 42; m <= 115; m++)
+      {
+        fout = ((double)(m) * fref)/((double)(n1) * (1 << n2));
+        fvco_goal = (double)clock * (double)(1 << n2)*fac;
+        fvco = fout * (double)(1 << n2);
+        if (!WITHIN(fvco, 0.995*fvco_goal, 1.005*fvco_goal))
+          continue;
+        if (!WITHIN(fvco, fmin, fmax))
+          continue;
+#if 0
+	printf("clock=%i m=%i n1=%i n2=%i\n ",clock,m,n1,n2);
+#endif
+        return (m - 2) | ((n1 - 2)<<8) | (n2 << 13) ;
+      }
+    }
+  }
+  printf("ET6000: Illegal clock\n");
+  return 0;
 }
