@@ -429,6 +429,9 @@ static int ms_init(void)
         m_wheel_delta = 0;
 
     /* Added O_NDELAY here. */
+/*
+    if(__svgalib_mouse_fd!=-1)close(__svgalib_mouse_fd);
+*/
     if ((__svgalib_mouse_fd = open(m_dev, O_RDWR | O_NDELAY )) < 0)
 	return -1;
 
