@@ -11,14 +11,13 @@ there are still the following problems:
   * no 24bit modes. (maybe the hardware does not support)
   * pageflipping (in threeDKit) does not work. 
   * no acceleration (is there a program that uses it anyway?).
-  
 */
 
 
 #include <stdlib.h>
 #include <stdio.h>		/* for printf */
 #include <string.h>		/* for memset */
-#include <sys/mman.h>		/* mmap */
+#include <sys/mman.h>		
 #include <fcntl.h>
 #include <math.h>
 #include <unistd.h>
@@ -65,7 +64,7 @@ static unsigned long MMIOBASE, LINEARBASE;
 
 
 enum {
-   NV3_chipset = 0
+   Riva128 = 0, Riva128ZX, RivaTNT
 };
 
 static void nv3_setpage(int page)
@@ -437,9 +436,14 @@ static int nv3_test(void)
      }
    }
    found=__svgalib_pci_find_vendor_vga(0x12d2,buf);
+   if (found) {
+      found=__svgalib_pci_find_vendor_vga(0x10de,buf);
+      if (_ioperm) iopl(0);
+      if(found)return 0;
+   };
    if (_ioperm) iopl(0);
-   if (found)return 0;
-   if (((buf[0]>>16)&0xffff)!=0x18)return 0;
+   if ((((buf[0]>>16)&0xffff)!=0x18)&&(((buf[0]>>16)&0xffff)!=0x20))return 0;
+   if((buf[0]>>16)==0x20)nv3_chiptype=RivaTNT; else nv3_chiptype=Riva128;
    MMIOBASE=buf[4]&0xffffff00;
    LINEARBASE=buf[5]&0xffffff00;
    nv3_init(0,0,0);
@@ -548,6 +552,30 @@ DriverSpecs __svgalib_nv3_driverspecs =
 
 static int nv3_init(int force, int par1, int par2)
 {   nv3_unlock();
+    
+    if(MMIOBASE==0) {
+       unsigned long buf[64];
+       int _ioperm=0,found;
+       
+       if (getenv("IOPERM") == NULL) {
+         _ioperm=1;
+         if (iopl(3) < 0) {
+           printf("svgalib: nv3: cannot get I/O permissions\n");
+           exit(1);
+         }
+       }
+       found=__svgalib_pci_find_vendor_vga(0x12d2,buf);
+       if (_ioperm) iopl(0);
+       if (found)return 0;
+       if (found) {
+          found=__svgalib_pci_find_vendor_vga(0x10de,buf);
+          if (_ioperm) iopl(0);
+          if(found)return 0;
+       };
+       if((buf[0]>>16)==0x20)nv3_chiptype=RivaTNT; else nv3_chiptype=Riva128;
+       MMIOBASE=buf[4]&0xffffff00;
+       LINEARBASE=buf[5]&0xffffff00;
+    };
     if (force) {
 	nv3_memory = par1;
 	nv3_chiptype = par2;
@@ -559,8 +587,18 @@ static int nv3_init(int force, int par1, int par2)
        int boot0;
        
        boot0=PFB_Read(BOOT_0);
-       if(boot0&0x20)nv3_memory=8192; else nv3_memory=1024<<(boot0&3); 
-       nv3_chiptype=0;
+       switch(nv3_chiptype){
+          case Riva128:
+          case Riva128ZX:
+                 if(boot0&0x20)nv3_memory=8192; else nv3_memory=1024<<(boot0&3); 
+                 if(nv3_memory==1024)nv3_memory=8192;
+                 break;
+          case RivaTNT:
+                 nv3_memory=2048<<(boot0&3); 
+                 if(nv3_memory==2048)nv3_memory=32768;
+                 break;
+       };
+
     };
     if (__svgalib_driver_report) {
 	printf("Using RIVA 128 driver, %iKB.\n",nv3_memory);
@@ -581,6 +619,10 @@ static int nv3_init(int force, int par1, int par2)
     cardspecs->mapHorizontalCrtc = nv3_map_horizontal_crtc;
     cardspecs->matchProgrammableClock=nv3_match_programmable_clock;
     __svgalib_driverspecs = &__svgalib_nv3_driverspecs;
+    __svgalib_banked_mem_base=0xa0000;
+    __svgalib_banked_mem_size=0x10000;
+    __svgalib_linear_mem_base=LINEARBASE;
+    __svgalib_linear_mem_size=nv3_memory*0x400;
     return 0;
 }
 
