@@ -5,7 +5,6 @@
 #include <stdlib.h>
 #include <stdio.h>		/* for printf */
 #include <string.h>		/* for memset */
-#include <sys/mman.h>		/* mmap */
 #include <unistd.h>
 #include "vga.h"
 #include "libvga.h"
@@ -31,8 +30,10 @@ static void vesa_unlock(void);
 
 static int vesa_memory,vesa_chiptype;
 static int vesa_is_linear, vesa_logical_width, vesa_bpp, vesa_granularity;
-static int vesa_regs_size;
+static int vesa_regs_size, vesa_linear_base, vesa_last_mode_set;
 static struct LRMI_regs vesa_r;
+static int vesa_read_write, vesa_read_window, vesa_write_window;
+static void * LRMI_mem1, * LRMI_mem2;
 
 static CardSpecs *cardspecs;
 static struct
@@ -49,43 +50,51 @@ vesa_r.eax=0x4f05;
 vesa_r.ebx=0;
 vesa_r.edx=page*64/vesa_granularity;
 __svgalib_LRMI_int(0x10,&vesa_r);
+
+if(vesa_read_write){
+   vesa_r.eax=0x4f05;
+   vesa_r.ebx=1;
+   vesa_r.edx=page*64/vesa_granularity;
+   __svgalib_LRMI_int(0x10,&vesa_r);
+};
 }
 
-/*
+
 static int __svgalib_vesa_inlinearmode(void)
 {
 return vesa_is_linear;
 }
-*/
+
 
 /* Fill in chipset specific mode information */
 
 static void vesa_getmodeinfo(int mode, vga_modeinfo *modeinfo)
 {
 
-     if(modeinfo->colors==16)return;
+    if(modeinfo->colors==16)return;
 
-	modeinfo->maxpixels = vesa_memory*1024/modeinfo->bytesperpixel;
+    modeinfo->maxpixels = vesa_memory*1024/modeinfo->bytesperpixel;
     modeinfo->maxlogicalwidth = 4088; /* just a guess, */
-	modeinfo->startaddressrange = vesa_memory * 1024 - 1;
-	modeinfo->haveblit = 0;
+    modeinfo->startaddressrange = vesa_memory * 1024 - 1;
+    modeinfo->haveblit = 0;
     modeinfo->flags &= ~HAVE_RWPAGE;
+    modeinfo->flags |= vesa_read_write;  /* sets HAVE_RWPAGE bit */
 
-/* for linear need VBE2 */
-/*    if (modeinfo->bytesperpixel >= 1) {
+    /* for linear need VBE2 */
+    if(vesa_chiptype==1)
+      if (modeinfo->bytesperpixel >= 1) {
 	modeinfo->flags |= CAPABLE_LINEAR;
         if (__svgalib_vesa_inlinearmode())
 	    modeinfo->flags |= IS_LINEAR;
-    }
-*/      
+      }
+      
 }
 
 /* Read and save chipset-specific registers */
-
 static int vesa_saveregs(unsigned char regs[])
 { 
   void * buf;
-  buf=__svgalib_LRMI_alloc_real(vesa_regs_size);
+  buf=LRMI_mem1;
   vesa_r.eax=0x4f04;
   vesa_r.ebx=0;
   vesa_r.es=((long)buf)>>4;
@@ -93,7 +102,6 @@ static int vesa_saveregs(unsigned char regs[])
   vesa_r.ecx=__svgalib_VESA_savebitmap;
   __svgalib_LRMI_int(0x10,&vesa_r);
   memcpy(&regs[VGA_TOTAL_REGS],buf,vesa_regs_size);  
-  __svgalib_LRMI_free_real(buf);
     return vesa_regs_size;
 }
 
@@ -103,7 +111,7 @@ static void vesa_setregs(const unsigned char regs[], int mode)
 {   
 
   void * buf;
-  buf=__svgalib_LRMI_alloc_real(vesa_regs_size);
+  buf=LRMI_mem1;
   memcpy(buf,&regs[VGA_TOTAL_REGS],vesa_regs_size);  
   vesa_r.eax=0x4f04;
   vesa_r.ebx=0;
@@ -111,7 +119,6 @@ static void vesa_setregs(const unsigned char regs[], int mode)
   vesa_r.edx=2;
   vesa_r.ecx=__svgalib_VESA_savebitmap;
   __svgalib_LRMI_int(0x10,&vesa_r);
-__svgalib_LRMI_free_real(buf);
 }
 
 
@@ -151,39 +158,43 @@ static int vesa_setmode(int mode, int prv_mode)
 {
     vesa_bpp=1;
     vesa_granularity=1;
-    if ((mode < G640x480x256)
-	|| (mode == G720x348x2)) {
+    if ((mode < G640x480x256)|| (mode == G720x348x2)) {
 
-    if(__svgalib_vesatext){
-        vesa_r.eax=0x4f02; /* make sure we are in a regular VGA mode before we start */
-        vesa_r.ebx=0x3;    /* without this, if we start in SVGA mode the result might */
-        __svgalib_LRMI_int(0x10,&vesa_r); /* be something weird */
-    };
+        if(__svgalib_vesatext){
+            vesa_r.eax=0x4f02; /* make sure we are in a regular VGA mode before we start */
+            vesa_r.ebx=0x3;    /* without this, if we start in SVGA mode the result might */
+            __svgalib_LRMI_int(0x10,&vesa_r); /* be something weird */
+        };
 	return __svgalib_vga_driverspecs.setmode(mode, prv_mode);
     }
-    if (!vesa_modeavailable(mode))
-	return 1;
-  vesa_r.eax=0x4f02;
-  vesa_r.ebx=SVGALIB_VESA[mode]|0x8000;
-  __svgalib_LRMI_int(0x10,&vesa_r);
+    if (!vesa_modeavailable(mode)) return 1;
+    vesa_r.eax=0x4f02;
+    vesa_r.ebx=SVGALIB_VESA[mode]|0x8000|(vesa_is_linear*0x4000);
+    vesa_last_mode_set=vesa_r.ebx;
+    __svgalib_LRMI_int(0x10,&vesa_r);
 
-	vesa_data.info = __svgalib_LRMI_alloc_real(sizeof(struct vbe_info_block)
-	 + sizeof(struct vbe_mode_info_block));
-	vesa_data.mode = (struct vbe_mode_info_block *)(vesa_data.info + 1);
-	vesa_r.eax = 0x4f01;
-        vesa_r.ecx=SVGALIB_VESA[mode];
-	vesa_r.es = (unsigned int)vesa_data.mode >> 4;
-	vesa_r.edi = (unsigned int)vesa_data.mode&0xf;    
-/*  memcpy(vesa_data.info->vbe_signature, "VBE2", 4);*/
-        __svgalib_LRMI_int(0x10, &vesa_r);
-        vesa_logical_width=vesa_data.mode->bytes_per_scanline;
-        vesa_bpp=(vesa_data.mode->bits_per_pixel+7)/8;
-        if(vesa_logical_width==0) vesa_logical_width=vesa_bpp*vesa_data.mode->x_resolution;
+    vesa_data.info = LRMI_mem2 ;
+    vesa_data.mode = (struct vbe_mode_info_block *)(vesa_data.info + 1);
+    vesa_r.eax = 0x4f01;
+    vesa_r.ecx=SVGALIB_VESA[mode];
+    vesa_r.es = (unsigned int)vesa_data.mode >> 4;
+    vesa_r.edi = (unsigned int)vesa_data.mode&0xf;    
+    __svgalib_LRMI_int(0x10, &vesa_r);
+    vesa_logical_width=vesa_data.mode->bytes_per_scanline;
+    vesa_bpp=(vesa_data.mode->bits_per_pixel+7)/8;
+    if(vesa_logical_width==0) vesa_logical_width=vesa_bpp*vesa_data.mode->x_resolution;
                                         	/* if not reported then guess */
-        vesa_granularity=vesa_data.mode->win_granularity;
-        if(vesa_granularity==0)vesa_granularity=64; /* if not reported then guess */
-        __svgalib_LRMI_free_real(vesa_data.info);
-
+    vesa_granularity=vesa_data.mode->win_granularity;
+    if(vesa_granularity==0)vesa_granularity=64; /* if not reported then guess */
+    if(vesa_chiptype==1)vesa_linear_base=vesa_data.mode->phys_base_ptr;
+    vesa_read_write=0;
+    vesa_read_window=0;
+    vesa_write_window=0;
+    if((vesa_data.mode->win_a_attributes&6)!=6){ 
+        vesa_read_write=1;
+        if ((vesa_data.mode->win_b_attributes&2) == 2) vesa_read_window=1;
+        if ((vesa_data.mode->win_b_attributes&4) == 4) vesa_write_window=1;
+    }
     return 0;
 }
 
@@ -209,17 +220,13 @@ static int vesa_test(void)
 {
 
         __svgalib_LRMI_init();
-	vesa_data.info = __svgalib_LRMI_alloc_real(sizeof(struct vbe_info_block)
-	 + sizeof(struct vbe_mode_info_block));
+	vesa_data.info = LRMI_mem2;
 	vesa_data.mode = (struct vbe_mode_info_block *)(vesa_data.info + 1);
 	vesa_r.eax = 0x4f00;
 	vesa_r.es = (unsigned int)vesa_data.info >> 4;
 	vesa_r.edi = 0;
 
-/*	memcpy(vbe.info->vbe_signature, "VBE2", 4);*/
-
         __svgalib_LRMI_int(0x10, &vesa_r);
-        __svgalib_LRMI_free_real(vesa_data.info);
        if (vesa_r.eax!=0x4f) return 0;
        return !vesa_init(0,0,0);
 }
@@ -228,9 +235,17 @@ static int vesa_test(void)
 /* No r/w paging */
 static void vesa_setrdpage(int page)
 {
+vesa_r.eax=0x4f05;
+vesa_r.ebx=vesa_read_window;
+vesa_r.edx=page*64/vesa_granularity;
+__svgalib_LRMI_int(0x10,&vesa_r);
 }
 static void vesa_setwrpage(int page)
 {
+vesa_r.eax=0x4f05;
+vesa_r.ebx=vesa_write_window;
+vesa_r.edx=page*64/vesa_granularity;
+__svgalib_LRMI_int(0x10,&vesa_r);
 }
 
 
@@ -261,8 +276,20 @@ static void vesa_setlogicalwidth(int width)
 
 static int vesa_linear(int op, int param)
 {
-if (op==LINEAR_ENABLE || op==LINEAR_DISABLE){ vesa_is_linear=1-vesa_is_linear; return 0;}
-if (op==LINEAR_QUERY_BASE) {return 0 ;}
+if (op==LINEAR_ENABLE) {  
+  vesa_r.eax=0x4f02;
+  vesa_r.ebx=vesa_last_mode_set|0x4000;
+  __svgalib_LRMI_int(0x10,&vesa_r);
+  vesa_is_linear=1;
+};
+
+if (op==LINEAR_DISABLE){ 
+  vesa_r.eax=0x4f02;
+  vesa_r.ebx=vesa_last_mode_set;
+  __svgalib_LRMI_int(0x10,&vesa_r);
+  vesa_is_linear=0;
+};
+if (op==LINEAR_QUERY_BASE) {return vesa_linear_base ;}
 if (op == LINEAR_QUERY_RANGE || op == LINEAR_QUERY_GRANULARITY) return 0;		/* No granularity or range. */
     else return -1;		/* Unknown function. */
 }
@@ -330,9 +357,9 @@ static int vesa_init(int force, int par1, int par2)
         vesa_memory=4096; 
     };
 
+    __svgalib_LRMI_init();
     for(i=0;i<__GLASTMODE;i++)SVGALIB_VESA[i]=((i<G640x480x256)||(i==G720x348x2));
 
-    __svgalib_LRMI_init();
 	vesa_data.info = __svgalib_LRMI_alloc_real(sizeof(struct vbe_info_block)
 	 + sizeof(struct vbe_mode_info_block));
 	vesa_data.mode = (struct vbe_mode_info_block *)(vesa_data.info + 1);
@@ -340,10 +367,12 @@ static int vesa_init(int force, int par1, int par2)
 	vesa_r.es = (unsigned int)vesa_data.info >> 4;
 	vesa_r.edi = 0;
     
-/*	memcpy(vbe.info->vbe_signature, "VBE2", 4);*/
+	memcpy(vesa_data.info->vbe_signature, "VBE2", 4);
 
         __svgalib_LRMI_int(0x10, &vesa_r);
     
+        if(vesa_data.info->vbe_version>=0x0200)vesa_chiptype=1 ; else vesa_chiptype=0;
+
         vesa_memory = vesa_data.info->total_memory*64;
         
 	mode_list = (short int *)(vesa_data.info->video_mode_list_seg * 16 + vesa_data.info->video_mode_list_off);
@@ -357,6 +386,9 @@ static int vesa_init(int force, int par1, int par2)
 		vesa_r.es = (unsigned int)vesa_data.mode >> 4;
 		vesa_r.edi = (unsigned int)vesa_data.mode & 0xf;
 
+                if((vesa_chiptype==1)&&
+                   (vesa_data.mode->mode_attributes&0x80))
+                   vesa_linear_base=vesa_data.mode->phys_base_ptr;
 		if (!__svgalib_LRMI_int(0x10, &vesa_r))
 			{
 			fprintf(stderr, "Can't get mode info (vm86 failure)\n");
@@ -364,18 +396,24 @@ static int vesa_init(int force, int par1, int par2)
 			}
 
 		if (vesa_data.mode->memory_model == VBE_MODEL_RGB)
-                   {  
+                                   {  
                       if((vesa_data.mode->rsvd_mask_size==8)||(vesa_data.mode->bits_per_pixel==32))
                          {  
                             switch(vesa_data.mode->y_resolution){
                                case 200: if(vesa_data.mode->x_resolution==320)
                                   SVGALIB_VESA[G320x200x16M32]=*mode_list; break;
-/*                               case 240: if(vesa_data.mode->x_resolution==320)
+                               case 240: if(vesa_data.mode->x_resolution==320)
                                   SVGALIB_VESA[G320x240x16M32]=*mode_list; break;
-*/                               case 480: if(vesa_data.mode->x_resolution==640)
+                               case 300: if(vesa_data.mode->x_resolution==400)
+                                  SVGALIB_VESA[G400x300x16M32]=*mode_list; break;
+                               case 384: if(vesa_data.mode->x_resolution==512)
+                                  SVGALIB_VESA[G512x384x16M32]=*mode_list; break;
+                               case 480: if(vesa_data.mode->x_resolution==640)
                                   SVGALIB_VESA[G640x480x16M32]=*mode_list; break;
                                case 600: if(vesa_data.mode->x_resolution==800)
                                   SVGALIB_VESA[G800x600x16M32]=*mode_list; break;
+                               case 720: if(vesa_data.mode->x_resolution==960)
+                                  SVGALIB_VESA[G960x720x16M32]=*mode_list; break;
                                case 768: if(vesa_data.mode->x_resolution==1024)
                                   SVGALIB_VESA[G1024x768x16M32]=*mode_list; break;
                                case 864: if(vesa_data.mode->x_resolution==1152)
@@ -384,18 +422,26 @@ static int vesa_init(int force, int par1, int par2)
                                   SVGALIB_VESA[G1280x1024x16M32]=*mode_list; break;
                                case 1200: if(vesa_data.mode->x_resolution==1600)
                                   SVGALIB_VESA[G1600x1200x16M32]=*mode_list; break;
+                               case 1440: if(vesa_data.mode->x_resolution==1920)
+                                  SVGALIB_VESA[G1920x1440x16M32]=*mode_list; break;
                             }
                          } else
                          {  i=0;
                             switch(vesa_data.mode->y_resolution){
                                case 200: if(vesa_data.mode->x_resolution==320)
                                   i=G320x200x32K; break;
-/*                               case 240: if(vesa_data.mode->x_resolution==320)
+                               case 240: if(vesa_data.mode->x_resolution==320)
                                   i=G320x240x32K; break;
-*/                               case 480: if(vesa_data.mode->x_resolution==640)
+                               case 300: if(vesa_data.mode->x_resolution==400)
+                                  i=G400x300x32K; break;
+                               case 384: if(vesa_data.mode->x_resolution==512)
+                                  i=G512x384x32K; break;
+                               case 480: if(vesa_data.mode->x_resolution==640)
                                   i=G640x480x32K; break;
                                case 600: if(vesa_data.mode->x_resolution==800)
                                   i=G800x600x32K; break;
+                               case 720: if(vesa_data.mode->x_resolution==960)
+                                  i=G960x720x32K; break;
                                case 768: if(vesa_data.mode->x_resolution==1024)
                                   i=G1024x768x32K; break;
                                case 864: if(vesa_data.mode->x_resolution==1152)
@@ -404,11 +450,13 @@ static int vesa_init(int force, int par1, int par2)
                                   i=G1280x1024x32K; break;
                                case 1200: if(vesa_data.mode->x_resolution==1600)
                                   i=G1600x1200x32K; break;
+                               case 1440: if(vesa_data.mode->x_resolution==1920)
+                                  i=G1920x1440x32K; break;
                                };
                             if(i>0)switch(vesa_data.mode->green_mask_size){
                                case 5:SVGALIB_VESA[i]=*mode_list; break;
                                case 6:SVGALIB_VESA[i+1]=*mode_list; break;
-                               case 8:SVGALIB_VESA[i+2]=*mode_list; break;
+                               case 8:if(vesa_data.mode->rsvd_mask_size==0)SVGALIB_VESA[i+2]=*mode_list; break;
                             };
                          };	
                 }
@@ -417,12 +465,22 @@ static int vesa_init(int force, int par1, int par2)
                     switch(vesa_data.mode->y_resolution){
                        case 200: if(vesa_data.mode->x_resolution==320)
                           SVGALIB_VESA[G320x200x256]=*mode_list; break;
-/*                       case 240: if(vesa_data.mode->x_resolution==320)
-                          SVGALIB_VESA[G320x240x256]=*mode_list; break;
-*/                       case 480: if(vesa_data.mode->x_resolution==640)
+                       case 240: if(vesa_data.mode->x_resolution==320) {
+                         if(vesa_data.mode->number_of_planes==8)
+                            SVGALIB_VESA[G320x240x256]=*mode_list; 
+                            else SVGALIB_VESA[G320x240x256V]=*mode_list; 
+                         }; 
+                         break;
+                       case 300: if(vesa_data.mode->x_resolution==400)
+                          SVGALIB_VESA[G400x300x256]=*mode_list; break;
+                       case 384: if(vesa_data.mode->x_resolution==512)
+                          SVGALIB_VESA[G512x384x256]=*mode_list; break;
+                       case 480: if(vesa_data.mode->x_resolution==640)
                           SVGALIB_VESA[G640x480x256]=*mode_list; break;
                        case 600: if(vesa_data.mode->x_resolution==800)
                           SVGALIB_VESA[G800x600x256]=*mode_list; break;
+                       case 720: if(vesa_data.mode->x_resolution==960)
+                          SVGALIB_VESA[G960x720x256]=*mode_list; break;
                        case 768: if(vesa_data.mode->x_resolution==1024)
                           SVGALIB_VESA[G1024x768x256]=*mode_list; break;
                        case 864: if(vesa_data.mode->x_resolution==1152)
@@ -431,6 +489,8 @@ static int vesa_init(int force, int par1, int par2)
                           SVGALIB_VESA[G1280x1024x256]=*mode_list; break;
                        case 1200: if(vesa_data.mode->x_resolution==1600)
                           SVGALIB_VESA[G1600x1200x256]=*mode_list; break;
+                       case 1440: if(vesa_data.mode->x_resolution==1920)
+                          SVGALIB_VESA[G1920x1440x256]=*mode_list; break;
                     }
                 }
 		else if (vesa_data.mode->memory_model == VBE_MODEL_PACKED)
@@ -438,12 +498,22 @@ static int vesa_init(int force, int par1, int par2)
                     if(vesa_data.mode->bits_per_pixel==8)switch(vesa_data.mode->y_resolution){
                        case 200: if(vesa_data.mode->x_resolution==320)
                           SVGALIB_VESA[G320x200x256]=*mode_list; break;
-/*                       case 240: if(vesa_data.mode->x_resolution==320)
-                          SVGALIB_VESA[G320x240x256]=*mode_list; break;
-*/                       case 480: if(vesa_data.mode->x_resolution==640)
+                       case 240: if(vesa_data.mode->x_resolution==320) {
+                         if(vesa_data.mode->number_of_planes==8)
+                            SVGALIB_VESA[G320x240x256]=*mode_list; 
+                            else SVGALIB_VESA[G320x240x256V]=*mode_list; 
+                         }; 
+                         break;
+                       case 300: if(vesa_data.mode->x_resolution==400)
+                          SVGALIB_VESA[G400x300x256]=*mode_list; break;
+                       case 384: if(vesa_data.mode->x_resolution==512)
+                          SVGALIB_VESA[G512x384x256]=*mode_list; break;
+                       case 480: if(vesa_data.mode->x_resolution==640)
                           SVGALIB_VESA[G640x480x256]=*mode_list; break;
                        case 600: if(vesa_data.mode->x_resolution==800)
                           SVGALIB_VESA[G800x600x256]=*mode_list; break;
+                       case 720: if(vesa_data.mode->x_resolution==960)
+                          SVGALIB_VESA[G960x720x256]=*mode_list; break;
                        case 768: if(vesa_data.mode->x_resolution==1024)
                           SVGALIB_VESA[G1024x768x256]=*mode_list; break;
                        case 864: if(vesa_data.mode->x_resolution==1152)
@@ -452,6 +522,8 @@ static int vesa_init(int force, int par1, int par2)
                           SVGALIB_VESA[G1280x1024x256]=*mode_list; break;
                        case 1200: if(vesa_data.mode->x_resolution==1600)
                           SVGALIB_VESA[G1600x1200x256]=*mode_list; break;
+                       case 1440: if(vesa_data.mode->x_resolution==1920)
+                          SVGALIB_VESA[G1920x1440x256]=*mode_list; break;
                     }
                 };
 		mode_list++;
@@ -481,8 +553,21 @@ static int vesa_init(int force, int par1, int par2)
     cardspecs->mapHorizontalCrtc = vesa_map_horizontal_crtc;
     cardspecs->matchProgrammableClock=vesa_match_programmable_clock;
     __svgalib_driverspecs = &__svgalib_vesa_driverspecs;
+
+    LRMI_mem1 = __svgalib_LRMI_alloc_real(vesa_regs_size);
+    LRMI_mem2 = __svgalib_LRMI_alloc_real(sizeof(struct vbe_info_block)
+                                    + sizeof(struct vbe_mode_info_block));
+
+    __svgalib_banked_mem_base=0xa0000;
+    __svgalib_banked_mem_size=0x10000;
+    if(vesa_chiptype==1) {
+      __svgalib_linear_mem_base=vesa_linear_base;
+      __svgalib_linear_mem_size=vesa_memory*0x400;
+    };
+    
     if (__svgalib_driver_report) {
-	printf("Using VESA driver, %iKB.\n",vesa_memory);
+	printf("Using VESA driver, %iKB. %s\n",vesa_memory,
+                vesa_chiptype?"VBE2.0":"VBE1.2");
     }
     return 0;
 }
